# Copyright 2021 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from enum import Enum

import pysam

from eva_vcf_merge.utils import are_all_elements_unique


class SampleSetType(Enum):
    # Same sample names in the same order, e.g. [(A, B), (A, B)]
    SINGLE_SET = 'single set'
    # Same sample names but in different orders, e.g. [(A, B), (B, A)]
    UNSORTED_SINGLE_SET = 'unsorted single set'
    # Distinct non-overlapping sample names, e.g. [(A, B), (C, D, E)]
    UNIQUE_SETS = 'unique sets'
    # Overlapping sample names, e.g. [(A, B), (C, D, A)]
    OVERLAPPING_SETS = 'overlapping sets'


class MergeType(Enum):
    # Horizontally mergeable VCFs have non-overlapping sample sets and can be combined to create
    # a single multi-sample file.
    HORIZONTAL = 'horizontal'
    # Vertically mergeable VCFs have identical sample sets and can be combined to create a single
    # multi-chromosome or multi-variant file.
    VERTICAL = 'vertical'


def detect_merge_type(vcf_files):
    """
    Detect what type of merge should be used for a list of VCFs.

    :param vcf_files: list of vcf filepaths
    :return: MergeType or None if it could not be detected
    """
    file_to_sample_names = {}
    # retrieve all the sample_names from the VCF files
    for file_path in vcf_files:
        file_to_sample_names[file_path] = get_samples_from_vcf(file_path)
    # Check that all the samples are the same and in the same order to enable horizontal merging
    sample_info = compare_sample_sets(file_to_sample_names.values())
    if sample_info == SampleSetType.UNIQUE_SETS:
        return MergeType.HORIZONTAL
    elif sample_info == SampleSetType.SINGLE_SET:
        return MergeType.VERTICAL
    return None


def get_samples_from_vcf(vcf_file):
    """Get the list of samples present in a single VCF file."""
    with pysam.VariantFile(vcf_file, 'r') as vcf_in:
        samples = list(vcf_in.header.samples)
    return samples


def compare_sample_sets(list_of_sample_names):
    """
    Compare sample names from different VCFs to determine whether they are identical, distinct, or overlapping.

    :param list_of_sample_names: list of list of sample names in each VCF
    :return: SampleSetType or None if it could not be detected
    """
    set_of_sample_names = set(tuple(s) for s in list_of_sample_names)
    set_of_sample_names_sorted = set([tuple(sorted(list(s))) for s in set_of_sample_names])
    if len(set_of_sample_names) == 1:
        return SampleSetType.SINGLE_SET
    elif len(set_of_sample_names_sorted) == 1:
        return SampleSetType.UNSORTED_SINGLE_SET
    elif are_all_elements_unique(itertools.chain(*list_of_sample_names)):
        return SampleSetType.UNIQUE_SETS
    elif len(set_of_sample_names_sorted) == len(list_of_sample_names):
        return SampleSetType.OVERLAPPING_SETS
    return None



# Copyright 2021 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

from ebi_eva_common_pyutils.nextflow import NextFlowPipeline, NextFlowProcess

from eva_vcf_merge.detect import MergeType
from eva_vcf_merge.multistage import get_multistage_merge_pipeline
from eva_vcf_merge.utils import get_valid_filename, validate_aliases


class VCFMerger:
    def __init__(self, bgzip_binary, bcftools_binary, nextflow_binary, nextflow_config, output_dir):
        self.bgzip_binary = bgzip_binary
        self.bcftools_binary = bcftools_binary
        self.nextflow_binary = nextflow_binary
        self.nextflow_config = nextflow_config
        self.output_dir = output_dir
        self.working_dir = os.path.join(output_dir, 'nextflow')

    def horizontal_merge(self, vcf_groups, chunk_size=500, resume=True):
        return self.common_merge(MergeType.HORIZONTAL, vcf_groups, chunk_size, resume)

    def vertical_merge(self, vcf_groups, chunk_size=500, resume=True):
        return self.common_merge(MergeType.VERTICAL, vcf_groups, chunk_size, resume)

    def common_merge(self, merge_type, vcf_groups, chunk_size=500, resume=True):
        """
        Merge groups of vcfs horizontally or vertically.

        :param merge_type: vertical or horizontal merge
        :param vcf_groups: dict mapping a string (e.g. an analysis alias) to a group of vcf files to be merged
        :param chunk_size: number of vcfs to merge at once (default 500)
        :param resume: whether to resume pipeline (default true)
        :returns: dict of merged filenames
        """
        if not validate_aliases(vcf_groups.keys()):
            raise ValueError('Aliases must be unique when converted to filenames')
        pipeline, merged_filenames = self.generate_merge_pipeline(merge_type, vcf_groups, chunk_size)
        workflow_file = os.path.join(self.working_dir, "merge.nf")

        os.makedirs(self.working_dir, exist_ok=True)
        pipeline.run_pipeline(
            workflow_file_path=workflow_file,
            working_dir=self.working_dir,
            nextflow_binary_path=self.nextflow_binary,
            nextflow_config_path=self.nextflow_config,
            resume=resume
        )
        # move merged files to output directory and rename with alias
        for alias in merged_filenames:
            safe_alias = get_valid_filename(alias)
            target_filename = os.path.join(self.output_dir, f'{safe_alias}_merged.vcf.gz')
            shutil.move(merged_filenames[alias], target_filename)
            shutil.move(merged_filenames[alias] + '.csi', target_filename + '.csi')
            shutil.move(merged_filenames[alias] + '.tbi', target_filename + '.tbi')
            merged_filenames[alias] = target_filename
        return merged_filenames

    def generate_merge_pipeline(self, merge_type, vcf_groups, chunk_size):
        """
        Generate merge pipeline, including compressing and indexing VCFs.

        :param merge_type: vertical or horizontal merge
        :param vcf_groups: dict mapping a string to a group of vcf files to be merged
        :param chunk_size: number of vcfs to merge at once
        :return: complete NextFlowPipeline and dict of merged filenames
        """
        full_pipeline = NextFlowPipeline()
        merged_filenames = {}
        for alias_idx, (alias, vcfs) in enumerate(vcf_groups.items()):
            compress_pipeline, compressed_vcfs = self.compress_and_index(alias_idx, vcfs)
            if merge_type == MergeType.HORIZONTAL:
                merge_pipeline, merged_filename = get_multistage_merge_pipeline(
                    alias=alias_idx,
                    vcf_files=compressed_vcfs,
                    chunk_size=chunk_size,
                    processing_dir=self.working_dir,
                    bcftools_binary=self.bcftools_binary,
                    process_name='merge',
                    process_command=self.merge_command
                )
            else:
                merge_pipeline, merged_filename = get_multistage_merge_pipeline(
                    alias=alias_idx,
                    vcf_files=compressed_vcfs,
                    chunk_size=chunk_size,
                    processing_dir=self.working_dir,
                    bcftools_binary=self.bcftools_binary,
                    process_name='concat',
                    process_command=self.concat_command
                )
            pipeline = NextFlowPipeline.join_pipelines(compress_pipeline, merge_pipeline)
            full_pipeline = NextFlowPipeline.join_pipelines(full_pipeline, pipeline)
            merged_filenames[alias] = merged_filename
        return full_pipeline, merged_filenames

    def compress_and_index(self, alias, vcfs):
        """
        Bgzip-compress and CSI-index VCFs.

        :param alias: name of group of vcf files (used to name Nextflow processes uniquely)
        :param vcfs: list of vcf files
        :return: NextFlow pipeline and list of final filenames
        """
        dependencies = {}
        index_processes = []
        compressed_vcfs = []
        for i, vcf in enumerate(vcfs):
            compress_process = None
            if not vcf.endswith('gz'):
                compress_process = NextFlowProcess(
                    process_name=f'compress_{alias}_{i}',
                    command_to_run=f'{self.bgzip_binary} -c {vcf} > {vcf}.gz'
                )
                vcf = f'{vcf}.gz'
            compressed_vcfs.append(vcf)
            index_process = NextFlowProcess(
                process_name=f'index_{alias}_{i}',
                command_to_run=f'{self.bcftools_binary} index -f -c {vcf}'
            )
            index_processes.append(index_process)
            # each file's index depends only on compress (if present)
            dependencies[index_process] = [compress_process] if compress_process else []
        return NextFlowPipeline(dependencies), compressed_vcfs

    def merge_command(self, files_to_merge_list, output_vcf_file):
        return (f'{self.bcftools_binary} merge --merge all --file-list {files_to_merge_list} --threads 3 '
                f'-O z -o {output_vcf_file}')

    def concat_command(self, files_to_merge_list, output_vcf_file):
        return (f'{self.bcftools_binary} concat --allow-overlaps --remove-duplicates --file-list {files_to_merge_list} '
                f'-O z -o {output_vcf_file}')



# Copyright 2021 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re


def are_all_elements_unique(elements):
    """Check if there are any repeated element in the list of elements. If yes return False otherwise return True."""
    unique_elements = set()
    for element in elements:
        if element in unique_elements:
            return False
        unique_elements.add(element)
    return True


def get_valid_filename(s):
    """Return string with characters not allowed in filenames replaced by underscore."""
    return re.sub(r'[^-.0-9a-zA-Z]+', '_', s)


def validate_aliases(aliases):
    """Checks that each alias remains unique when converted to a valid filename."""
    return len(set(get_valid_filename(s) for s in aliases)) == len(aliases)



# Copyright 2021 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os

from ebi_eva_common_pyutils.nextflow import NextFlowPipeline, NextFlowProcess


def get_multistage_merge_pipeline(
        alias,
        vcf_files,
        processing_dir,
        chunk_size,
        bcftools_binary,
        process_name,
        process_command,
        stage=0,
        prev_stage_processes=None,
        pipeline=None
):
    """
    Generate Nextflow pipeline for multi-stage VCF merging or concatenation.

    # As an example, below is multi-stage VCF concatenation of 5 VCF files with 2-VCFs concatenated at a time (CHUNK_SIZE=2)
    # For illustration purposes only. Usually the CHUNK_SIZE is much higher (ex: 500).
    #
    #		    vcf1		    vcf2		vcf3		    vcf4		vcf5
    #               \		     /		       \		      /
    # Stage0:	     \		   /		        \		    /
    # -------	      \	     /		             \	      /
    #	    	vcf1_2=concat(vcf1,vcf2)	vcf3_4=concat(vcf3,vcf4)	vcf5    <---- 3 batches of concat in stage 0
    #		 		    \ 		                /
    # Stage1:	  		 \		              /
    # -------	   		  \	                /
    #				vcf1_2_3_4=concat(vcf1_2,vcf3_4)		            vcf5    <---- 2 batches of concat in stage 1
    #	 					      \ 		                            /
    # Stage2:	  		 		   \		 	                      /
    # -------	   		  			\	                            /
    #						      vcf1_2_3_4_5=concat(vcf1_2_3_4,vcf5)          <----- Final result
    """
    if not pipeline:
        pipeline = NextFlowPipeline()
        prev_stage_processes = []
    # If we are left with only one file, this means we have reached the last merge stage
    if len(vcf_files) == 1:
        # Add last indexing using tbi
        index_process = NextFlowProcess(
            process_name=f"final_index_tbi_{alias}",
            command_to_run=f"{bcftools_binary} index --tbi {vcf_files[0]}"
        )
        pipeline.add_dependencies({index_process: prev_stage_processes[-1:]})
        return pipeline, vcf_files[0]

    num_batches_in_stage = math.ceil(len(vcf_files) / chunk_size)
    curr_stage_processes = []
    output_vcf_files_from_stage = []
    for batch in range(0, num_batches_in_stage):
        # split files in the current stage into chunks based on chunk_size
        files_in_batch = vcf_files[(chunk_size * batch):(chunk_size * (batch + 1))]
        files_to_merge_list = write_files_to_merge_list(process_name, alias, files_in_batch, stage, batch, processing_dir)
        output_vcf_file = get_output_vcf_file_name(process_name, alias, stage, batch, processing_dir)

        # separate merge & index processes
        merge_process = NextFlowProcess(
            process_name=f"{process_name}{alias}_stage{stage}_batch{batch}",
            command_to_run=process_command(files_to_merge_list, output_vcf_file)
        )
        index_process = NextFlowProcess(
            process_name=f"index{alias}_stage{stage}_batch{batch}",
            command_to_run=f"{bcftools_binary} index --csi {output_vcf_file}"
        )
        # index depends only on this batch's merge
        pipeline.add_dependencies({index_process: [merge_process]})
        # next stage requires indexing to be complete from this stage
        curr_stage_processes.append(index_process)
        output_vcf_files_from_stage.append(output_vcf_file)

        # Merge batch in a given stage will have to wait until the completion of
        # n batches in the previous stage where n = chunk_size
        # Ex: In the illustration above stage 1/batch 0 depends on completion of stage 0/batch 0 and stage 0/batch 1
        # While output of any n batches from the previous stage can be worked on as they become available,
        # having a predictable formula simplifies pipeline generation and troubleshooting
        prev_stage_dependencies = prev_stage_processes[(chunk_size * batch):(chunk_size * (batch + 1))]
        pipeline.add_dependencies({merge_process: prev_stage_dependencies})
    prev_stage_processes = curr_stage_processes

    return get_multistage_merge_pipeline(
        alias, output_vcf_files_from_stage,
        processing_dir, chunk_size,
        bcftools_binary,
        process_name,
        process_command,
        stage=stage+1,
        prev_stage_processes=prev_stage_processes,
        pipeline=pipeline
    )


def write_files_to_merge_list(process_name, alias, files_to_merge, stage, batch, processing_dir):
    """
    Write the list of files to be merged for a given stage and batch
    """
    output_dir = get_output_dir(process_name, alias, stage, processing_dir)
    list_filename = os.path.join(output_dir, f"batch{batch}_files.list")
    os.makedirs(os.path.dirname(list_filename), exist_ok=True)
    with open(list_filename, "w") as handle:
        for filename in files_to_merge:
            handle.write(filename + "\n")
    return list_filename


def get_output_dir(process_name, alias, stage_index, processing_dir):
    """
    Get the file name with the list of files to be merged for a given stage and batch in the merge process
    """
    return os.path.join(processing_dir, f'{process_name}_{alias}', f"stage_{stage_index}")


def get_output_vcf_file_name(process_name, alias, stage_index, batch_index, processing_dir):
    return os.path.join(get_output_dir(process_name, alias, stage_index, processing_dir),
                        f"{process_name}{alias}_output_stage{stage_index}_batch{batch_index}.vcf.gz")



from distutils.core import setup
from os.path import join, abspath, dirname

base_dir = abspath(dirname(__file__))
requirements_txt = join(base_dir, 'requirements.txt')
requirements = [l.strip() for l in open(requirements_txt) if l and not l.startswith('#')]

version = open(join(base_dir, 'eva_vcf_merge', 'VERSION')).read().strip()

setup(
    name='eva_vcf_merge',
    packages=['eva_vcf_merge'],
    package_data={'eva_vcf_merge': ['VERSION']},
    version=version,
    license='Apache',
    description='EBI EVA - VCF merge library',
    url='https://github.com/EBIVariation/eva-vcf-merge',
    keywords=['ebi', 'eva', 'python', 'vcf'],
    install_requires=requirements,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ]
)
